/*
 *
 *  This file is part of
 *	MakeIndex - A formatter and format independent index processor
 *
 *  Copyright (C) 1989 by Chen & Harrison International Systems, Inc.
 *  Copyright (C) 1988 by Olivetti Research Center
 *  Copyright (C) 1987 by Regents of the University of California
 *
 *  Author:
 *	Pehong Chen
 *	Chen & Harrison International Systems, Inc.
 *	Palo Alto, California
 *	USA
 *	(phc@renoir.berkeley.edu or chen@orc.olivetti.com)
 *
 *  Contributors:
 *	Please refer to the CONTRIB file that comes with this release
 *	for a list of people who have contributed to this and/or previous
 *	release(s) of MakeIndex.
 *
 *  All rights reserved by the copyright holders.  See the copyright
 *  notice distributed with this software for a complete description of
 *  the conditions under which it is made available.
 *
 */

#define MKIND_C	1
#include "mkind.h"
#undef MKIND_C

FILE		*stddbg;
int		letter_ordering = FALSE;
int		compress_blanks = FALSE;
int		debug = FALSE;
int		merge_page = TRUE;
int		init_page = FALSE;
int		even_odd = -1;
int		verbose = TRUE;
int		german_sort = FALSE;
int		idx_dot = TRUE; /* flag which shows dot in ilg being active */
entry_number	idx_tt = 0;		/* total entry count (all files) */
entry_number	idx_et = 0;		/* erroneous entry count (all files) */
entry_number	idx_gt = 0;		/* good entry count (all files) */

FIELD_PTR	*idx_key;
FILE		*sty_fp;
FILE		*idx_fp;
FILE		*ind_fp;
FILE		*ilg_fp;

char		pageno[NUMBER_MAX];
const char	*sty_fn;
const char	*idx_fn;
const char	*ilg_fn;
const char	*ind_fn;
const char	*pgm_fn;

static int	fn_no = -1;		/* total number of files */
static const char *log_fn;
static FILE	*log_fp;
static char	base[FILENAME_MAX];
static int	need_version = TRUE;

#if    OS_PCDOS
#if    IBM_PC_TURBO
unsigned	_stklen = 0xf000;	/* Turbo C ignores size set in
					.exe file by LINK or EXEMOD,
					sigh... */
#endif /* IBM_PC_TURBO */
#endif /* OS_PCDOS */

#if OS_ATARI
long    _stksize = 20000L;		/* make the stack larger than 2KB */
					/* size must be given */
#endif /* OS_ATARI */

long totmem = 0L;			/* for debugging memory usage */

static	void	check_all ARGS((char *fn,int ind_given,int ilg_given,
				int log_given));
static	void	check_idx ARGS((char *fn,int open_fn));
static	void	chknew ARGS((size_t size));
static	void	find_pageno ARGS((void));
static	void	open_sty ARGS((char *fn));
static	void	prepare_idx ARGS((void));
static	void	process_idx ARGS((char * *fn,int use_stdin,int sty_given,
				  int ind_given,int ilg_given,int log_given));

int
#if STDC
main(int argc, char *argv[])
#else
main(argc, argv)
int     argc;
char   *argv[];
#endif
{
    char   **fns;			/* array of pointers to filenames */
    char   *ap;
    char   *p;
    int     use_stdin = FALSE;
    int     sty_given = FALSE;
    int     ind_given = FALSE;
    int     ilg_given = FALSE;
    int     log_given = FALSE;

    stddbg = stderr;			/* default debug output stream; */
					/* cannot be set at compile time */
					/* in some C implementations */

    /* determine program name */
#if (OS_ATARI | OS_VAXVMS | OS_BS2000 | OS_MVSXA | OS_VMCMS)
    p = "MakeIndex";			/* Use symbolic name on some systems */
#else
    p = strrchr(*argv, DIR_DELIM);
    if (p == NULL)
	p = *argv;
    else
	p++;

#if OS_PCDOS
    {
	register char *ext = p + strlen(p)-4;
	if ( *ext == '.' )
	    *ext = NUL;			/* cut off ".EXE" */
    }
    (void)strlwr(p);			/* lower program name    */
#endif /* OS_PCDOS */

#endif /* (OS_ATARI | OS_VAXVMS | OS_BS2000 | OS_MVSXA | OS_VMCMS) */

    pgm_fn = (const char*)p;

    fns = (char**)chkalloc(argc*sizeof(char*),sizeof(char*));

    /* process command line options */
    while (--argc > 0) {
	if (**++argv == SW_PREFIX) {
	    if (*(*argv + 1) == NUL)
		break;
	    for (ap = ++*argv; *ap != NUL; ap++)
	    {
		switch (OPTION_CHAR(*ap))
		{
		case 'c':		/* compress blanks */
		    compress_blanks = TRUE;
		    break;

		case 'd':	/* debugging (successive uses toggle) */
		    debug = TRUE;
		    break;

		case 'e':		/* redirect stderr and stddbg */
		    {
			argc--;
			if (argc <= 0)
			    FATAL("Expected -e <filename>\n%s","");
			++argv;
			if (STREQ(*argv,"-")) /* - puts stderr on stdout */
			    *stderr = *stdout;
			else
			{
			    if (REOPEN_OUT(*argv,stderr) == (FILE*)NULL)
			    {
				*stderr = *stdout; /* stderr was closed */
				FATAL("Cannot open error file [%s]\n",*argv);
			    }
			}
			stddbg = stderr; /* someday, this may be separately */
					/* redirectable */
		    }
		    break;

		case 'g':		/* enable german sort */
		    german_sort = TRUE;
		    break;

		case 'i':		/* use standard input */
		    use_stdin = TRUE;
		    break;

		case 'l':		/* enable letter ordering */
		    letter_ordering = TRUE;
		    break;

		case 'o':		/* output index file name */
		    argc--;
		    if (argc <= 0)
			FATAL("Expected -o <ind>\n%s","");
		    ind_fn = *++argv;
		    ind_given = TRUE;
		    break;

		case 'p':		/* initial page */
		    argc--;
		    if (argc <= 0)
			FATAL("Expected -p <num>\n%s","");
		    strcpy(pageno, *++argv);
		    init_page = TRUE;
		    if (STREQ(pageno, EVEN)) {
			log_given = TRUE;
			even_odd = 2;
		    } else if (STREQ(pageno, ODD)) {
			log_given = TRUE;
			even_odd = 1;
		    } else if (STREQ(pageno, ANY)) {
			log_given = TRUE;
			even_odd = 0;
		    }
		    break;

		case 'q':	/* supress progress message -- quiet mode */
		    verbose = FALSE;
		    break;

		case 'r':		/* disable range merge */
		    merge_page = FALSE;
		    break;

		case 's':		/* style file */
		    argc--;
		    if (argc <= 0)
			FATAL("Expected -s <stylefile>\n%s","");
		    open_sty(*++argv);
		    sty_given = TRUE;
		    break;

		case 't':		/* transcript file name */
		    argc--;
		    if (argc <= 0)
			FATAL("Expected -t <logfile>\n%s","");
		    ilg_fn = *++argv;
		    ilg_given = TRUE;
		    break;

		default:		/* bad option */
		    FATAL("Unknown option -%c.\n", *ap);
		    break;
		} /* end switch (OPTION_CHAR(*ap)) */
	    } /* end for (ap = ++*argv; *ap != NUL; ap++) */
	}
	else
	{
	    check_idx(*argv, FALSE);
	    fns[++fn_no] = *argv;
	}
    } /* end while (--argc > 0)  */

/* changes for 2.12 (May 20, 1993) by Julian Reschke (jr@ms.maus.de):
   if only one input file and no explicit style file was given */
#if OS_BS2000
/* not supported */
#else
	if (fn_no == 0 && !sty_given)
	{
		char temp[FILENAME_MAX + sizeof(INDEX_STY)];

		/* base set by last call to check_idx */
		(void)sprintf (temp, "%s%s", base, INDEX_STY);
		if (0 == access(temp, R_OK)) {
			open_sty (temp);
			sty_given = TRUE;
		}
	}
#endif

    if (german_sort)			/* change default quote character */
	idx_quote = IDX_QUOTE_GERMAN;

    process_precedence();	/* initialize default page precedence */
    process_idx(fns, use_stdin, sty_given, ind_given, ilg_given, log_given);
    idx_gt = idx_tt - idx_et;
    ALL_DONE;
    if (idx_gt > 0) {
	prepare_idx();
	sort_idx();
	gen_ind();
	MESSAGE("Output written in %s.\n", ind_fn);
    } else
	MESSAGE("Nothing written in %s.\n", ind_fn);

    MESSAGE("Transcript written in %s.\n", ilg_fn);
    chkfree();				/* free dynamic chunk space */
    CLOSE(ind_fp);
    CLOSE(ilg_fp);
    EXIT(((idx_et == 0) && (sty_ec == 0)) ? 0 : 1);

    return (0);				/* never executed--avoids complaints */
					/* about no return value */
}

/*--------------------------------------------------------------------*/

/***********************************************************************
malloc() implementations themselves usually allocate large chunks of
memory of fixed size in which they suballocate requests with
additional overhead of pointers, lengths, and alignment requirements.
Experiments with the program in tstmall.c indicate a typical policy of
aligning the request to a double-word boundary, then adding space for
two extra words.  Thus, allocation of a single byte uses up 24 bytes!
When a large chunk is used up, another large chunk is allocated with
an additional overhead of two more words.  Thus, we should allocate
our own chunks in sizes of malloc() large chunk sizes, less the
two-word/chunk malloc() overhead, and less our chunk pointer size.  On
some systems, optimal choice of CHUNKSIZE is critical if space is not
to be seriously wasted; a value of 2048 is best for DECstation ULTRIX
4.3.
***********************************************************************/

#ifndef CHUNKSIZE		/* can specify at compile time */
#define CHUNKSIZE	2048	/* suitable for most systems */

#if defined(__alpha) && defined(__unix__)
#undef CHUNKSIZE
#define CHUNKSIZE	2040	/* for DEC Alpha OSF/1 2.x */
#endif

#if defined(__host_mips) && defined(_SYSTYPE_SYSV) && !defined(__sgi)
#undef CHUNKSIZE
#define CHUNKSIZE	16892	/* weird size for MIPS RC6280 EP/IX 2.1.1 */
#endif

#endif /* CHUNKSIZE */

#ifndef MAXCHUNK
#define MAXCHUNK	(CHUNKSIZE - 2*sizeof(void*) - 2*sizeof(void*))
#endif

typedef struct KCHUNK
{
    struct KCHUNK	*prev;		/* pointer to previous chunk */
    char		*end; /* address of byte following end of storage[] */
    char 		storage[1];	/* space for this chunk */
} CHUNK, *CHUNK_PTR;

static CHUNK_PTR chkhead = (CHUNK_PTR)NULL; /* pointer to chain of chunks */
static char	*chknext = (char*)NULL;	/* next available in current chunk */
static long	chktotal = 0L;		/* total bytes allocated */


void *
#if STDC
chkalloc(size_t size, size_t alignment_size)
#else
chkalloc(size,alignment_size)	/* get size bytes with specified alignment */
size_t size;
size_t alignment_size;
#endif
{
    char *p;

    if (chkhead == (CHUNK_PTR)NULL)	/* then very first call */
	chknew(size);
    if (alignment_size < 1)
	alignment_size = 1;
    chknext = (char*)((long)(chknext + alignment_size - 1) &
		      ~(alignment_size - 1));
    if ((chknext + size) > chkhead->end)
	chknew(size);

    p = chknext;
    chknext += size;
    return ((void*)p);
}


void
chkfree(VOID_ARG)		/* free all dynamically-allocated chunks */
{
    CHUNK_PTR old_chkhead;

    MESSAGE("Dynamic memory allocated: %ld bytes.\n", chktotal);
    for ( ;  chkhead != (CHUNK_PTR)NULL;  )
    {
	old_chkhead = chkhead;
	chkhead = chkhead->prev;
	FREE(old_chkhead);
    }
    chktotal = 0L;
}


#ifdef MAX
#undef MAX
#endif
#define MAX(a,b)	( ((a) > (b)) ? (a) : (b) )

static void
#if STDC
chknew(size_t size)			/* get a new chunk from malloc() */
#else
chknew(size)
size_t size;
#endif
{
    size_t chunk_size;
    CHUNK_PTR new_chkhead;

    chunk_size = MAX(size,MAXCHUNK) + sizeof(CHUNK_PTR) + sizeof(char*);

    /* Here is the ONLY malloc() call in MakeIndex 2.13 (or later) */
    new_chkhead = (CHUNK_PTR)malloc(chunk_size);

    if (new_chkhead == (CHUNK_PTR)NULL)
    {
	/* When memory is exhausted, tell the user how far we got before
	   quitting. */
	MESSAGE("Current input file  = %s\n", idx_fn);
	MESSAGE("Current line number = %ld\n", (long)idx_lc);
	MESSAGE("Current entry total = %ld\n", (long)idx_tt);
	FATAL("Not enough core...abort.\n%s", "");
    }

    chktotal += chunk_size;		/* track total bytes allocated */
    (void)memset(new_chkhead,-1,chunk_size); /* for debugging only */

    new_chkhead->prev = chkhead;	/* attach to head of chunk chain */
    chkhead = new_chkhead;
    chkhead->end = (char*)chkhead + (int)chunk_size;
					/* 1 beyond end of storage[] */
    chknext = &(chkhead->storage[0]);
}


/*--------------------------------------------------------------------*/

static void
prepare_idx(VOID_ARG)
{
    NODE_PTR		ptr = head;
    entry_number	i = 0;

    if (debug)
    {
	totmem += idx_gt * sizeof(FIELD_PTR);
	(void)fprintf(stddbg,
		"\nDEBUG: prepare_idx(): chkalloc(%u,%u)\ttotmem = %ld\n",
		      (unsigned int)(idx_gt*sizeof(FIELD_PTR)),
		      (unsigned int)sizeof(void*), totmem);
    }

    if (head == (NODE_PTR)NULL)
	FATAL("No valid index entries collected.\n%s", "");

    idx_key = (FIELD_PTR *)chkalloc(idx_gt*sizeof(FIELD_PTR), sizeof(void*));
    for (i = 0; i < idx_gt; i++) {
	idx_key[i] = &(ptr->data);
	ptr = ptr->next;
    }
}


static void
#if STDC
process_idx(char *fn[], int use_stdin, int sty_given, int ind_given,
	    int ilg_given, int log_given)
#else
process_idx(fn, use_stdin, sty_given, ind_given, ilg_given, log_given)
char   *fn[];
int     use_stdin;
int     sty_given;
int     ind_given;
int     ilg_given;
int     log_given;
#endif
{
    int     i;
    static char Stdin[] = "stdin";
    static char Stdout[] = "stdout";
    static char Stderr[] = "stderr";

    if (fn_no == -1)
	/* use stdin if no input files specified */
	use_stdin = TRUE;
    else {
	check_all(fn[0], ind_given, ilg_given, log_given);
	PUT_VERSION;
	if (sty_given)
	    scan_sty();
	if (debug)
	    dump_sty();
	scan_idx();
	ind_given = TRUE;
	ilg_given = TRUE;
	for (i = 1; i <= fn_no; i++) {
	    check_idx(fn[i], TRUE);
	    scan_idx();
	}
    }

    if (use_stdin) {
	idx_fn = &Stdin[0];
	idx_fp = stdin;

	if (ind_given) {
	    if (!ind_fp && ((ind_fp = OPEN_OUT(ind_fn)) == NULL))
		FATAL("Can't create output index file %s.\n", ind_fn);
	} else {
	    ind_fn = &Stdout[0];
	    ind_fp = stdout;
	}

	if (ilg_given) {
	    if (!ilg_fp && ((ilg_fp = OPEN_OUT(ilg_fn)) == NULL))
		FATAL("Can't create transcript file %s.\n", ilg_fn);
	} else {
	    ilg_fn = &Stderr[0];
	    ilg_fp = stderr;
	    verbose = FALSE;		/* prevent duplicate output! */
	}

	if ((fn_no == -1) && sty_given)
	    scan_sty();
	if (debug)
	    dump_sty();
	if (need_version)
	    PUT_VERSION;
	scan_idx();
	fn_no++;
    }
}


static void
#if STDC
check_idx(char *fn, int open_fn)
#else
check_idx(fn, open_fn)
char   *fn;
int     open_fn;
#endif
{
    char	*ptr = fn;
    char	*ext;
    STATIC char	temp[FILENAME_MAX];
    int		with_ext = FALSE;
    int		i = 0;

#if OS_BS2000
#define WORK	":w:"
#define CATID	':'

    char	arg[sizeof(WORK) + 2 + 1]; /* argument for (void)sprintf() */
    STATIC char	message[3*FILENAME_MAX];	/* error message */
    static char	widx_fn[FILENAME_MAX + sizeof(WORK)];
    static FILE	*widx_fp;

    ext = strrchr(fn, EXT_DELIM);
    if ((ext != NULL) &&
	(((*fn == CATID) && (*(fn + 3) != DIR_DELIM)) ||
	 (((*fn == DIR_DELIM) || (*fn == CATID)) &&
	  (strchr(fn, EXT_DELIM) < ext)))) {
#else /* NOT OS_BS2000 */
    ext = strrchr(fn, EXT_DELIM);
    if ((ext != NULL) && (ext != fn) && (*(ext + 1) != DIR_DELIM)) {
#endif /* OS_BS2000 */
	with_ext = TRUE;
	while ((ptr != ext) && (i < sizeof(base)))
	    base[i++] = *ptr++;
    } else
	while ((*ptr != NUL) && (i < sizeof(base)))
	    base[i++] = *ptr++;

    if (i < sizeof(base))
	base[i] = NUL;
    else
	FATAL2("Index file name %s too long (max %u).\n",
	       base, (unsigned int)sizeof(base));

    idx_fn = fn;

#if OS_BS2000			       /* Search both on HOME-PVS and
					* WORK-PVS */
    (void)sprintf(widx_fn, "%s%s", WORK, idx_fn);
    if (((((idx_fp = OPEN_IN(idx_fn)) == NULL) && open_fn) ||
	 !(open_fn || idx_fp)) &&
	((((widx_fp = OPEN_IN(widx_fn)) == NULL) && open_fn) ||
	 !(open_fn || widx_fp)))
	if (with_ext) {
	    (void)sprintf(message, "Input index file %s%%s not found.\n",
		    (strchr(idx_fn, ':') || (strlen(WORK) == 0)) ?
		     NIL : "<" WORK ">");
	    FATAL(message, idx_fn);
	} else {

	    if (debug)
	    {
		totmem += FILENAME_MAX;
		(void)fprintf(stddbg,
		    "\nDEBUG: check_idx()-1: chkalloc(%u,%u)\ttotmem = %ld\n",
			      (unsigned int)FILENAME_MAX,
			      (unsigned int)sizeof(char),totmem);
	    }

	    idx_fn = (char *) chkalloc((size_t)FILENAME_MAX,sizeof(char));
	    (void)sprintf(idx_fn, "%s%s", base, INDEX_IDX);
	    (void)sprintf(widx_fn, "%s%s%s", WORK, base, INDEX_IDX);
	    if (((((idx_fp = OPEN_IN(idx_fn)) == NULL) && open_fn) ||
		 !(open_fn || idx_fp)) &&
		((((widx_fp = OPEN_IN(widx_fn)) == NULL) && open_fn) ||
		 !(open_fn || widx_fp))) {
		(void)sprintf(arg, "%s",
			(strchr(idx_fn, ':') || (strlen(WORK) == 0)) ?
			 NIL : "<" WORK ">");
		(void)sprintf(message,
			"Couldn't find either input file %s%%s nor %s%%s.\n",
			arg, arg);
		FATAL2(message, base, idx_fn);
	    }
	}
    if (!open_fn) {		       /* Close files if open */
	if (idx_fp)
	    FCLOSE(idx_fp);
	if (widx_fp)
	    FCLOSE(widx_fp);
    }
    if (open_fn && (!idx_fp) && widx_fp) {	/* File was found on PVS W */
	idx_fp = widx_fp;
	idx_fn = widx_fn;
    }
#else  /* NOT OS_BS2000 */
    if ( ( open_fn &&
	 ((idx_fp = OPEN_IN(idx_fn)) == NULL)
	 ) ||
	((!open_fn) && (access(idx_fn, R_OK) != 0)))
	if (with_ext) {
	    FATAL("Input index file %s not found.\n", idx_fn);
	} else {

	    if (debug)
	    {
		totmem += FILENAME_MAX;
		(void)fprintf(stddbg,
		    "\nDEBUG: check_idx()-2: chkalloc(%u,%u)\ttotmem = %ld\n",
			      (unsigned int)FILENAME_MAX,
			      (unsigned int)sizeof(char),totmem);
	    }

	    (void)sprintf(temp, "%s%s", base, INDEX_IDX);
	    idx_fn = (const char*)strcpy((char*)chkalloc(strlen(temp)+1,
							 sizeof(char)),temp);
	    if ((open_fn && ((idx_fp = OPEN_IN(idx_fn)) == NULL)) ||
		((!open_fn) && (access(idx_fn, R_OK) != 0))) {
		FATAL2("Couldn't find input index file %s or %s.\n", base,
		       idx_fn);
	    }
	}
#endif /* OS_BS2000 */
}


static void
#if STDC
check_all(char *fn, int ind_given, int ilg_given, int log_given)
#else
check_all(fn, ind_given, ilg_given, log_given)
char   *fn;
int     ind_given;
int     ilg_given;
int     log_given;
#endif
{
    STATIC char temp[FILENAME_MAX];

    check_idx(fn, TRUE);

    /* index output file */
    if (!ind_given) {
	(void)sprintf(temp, "%s%s", base, INDEX_IND);
	ind_fn = (const char*)strcpy((char*)chkalloc(strlen(temp)+1,
						     sizeof(char)),temp);
#if OS_BS2000			       /* Look on HOME-, then WORK-PVS */
	if ((ind_fp = OPEN_IN(ind_fn)) == NULL) {
	    (void)sprintf(temp, "%s%s%s", strchr(base, ':') ?
				   NIL : WORK, base, INDEX_IND);
	    ind_fn = (const char*)strcpy((char*)chkalloc(strlen(temp)+1,
							 sizeof(char)),temp);
	} else
	    CLOSE(ind_fp);
#endif /* OS_BS2000 */
    }
    if ((ind_fp = OPEN_OUT(ind_fn)) == NULL)
	FATAL("Can't create output index file %s.\n", ind_fn);

    /* index transcript file */
    if (!ilg_given) {
	(void)sprintf(temp, "%s%s", base, INDEX_ILG);
	ilg_fn = (const char*)strcpy((char*)chkalloc(strlen(temp)+1,
						     sizeof(char)),temp);
#if OS_BS2000			       /* Look on HOME-, then WORK-PVS */
	if ((ilg_fp = OPEN_IN(ilg_fn)) == NULL) {
	    (void)sprintf(temp, "%s%s%s", strchr(base, ':') ?
				   NIL : WORK, base, INDEX_ILG);
	    ilg_fn = (const char*)strcpy((char*)chkalloc(strlen(temp)+1,
							 sizeof(char)),temp);
	} else
	    CLOSE(ilg_fp);
#endif /* OS_BS2000 */
    }
    if ((ilg_fp = OPEN_OUT(ilg_fn)) == NULL)
	FATAL("Can't create transcript file %s.\n", ilg_fn);

#if OS_BS2000			       /* Search both on HOME-PVS ans
					* WORK-PVS */
    if (log_given) {
	(void)sprintf(temp, "%s%s%s", strchr(base, ':') ? NIL : WORK,
		base, INDEX_LOG);
	log_fn = (const char*)strcpy((char*)chkalloc(strlen(temp)+1,
						     sizeof(char)),temp);
	if ((log_fp = OPEN_IN(log_fn)) == NULL) {
	    if (strchr(base, ':'))
		(void)sprintf(temp, "%s%s", strrchr(base, ':') + 1,
			      INDEX_LOG);
	    else
		(void)sprintf(temp, "%s%s", base, INDEX_LOG);
	    log_fn = (const char*)strcpy((char*)chkalloc(strlen(temp)+1,
							 sizeof(char)),temp);
	    if ((log_fp = OPEN_IN(log_fn)) == NULL)
		FATAL2("Source log file %s%s not found.\n", WORK, log_fn);
	} else {
	    find_pageno();
	    CLOSE(log_fp);
	}
    }
#else
    if (log_given) {
	(void)sprintf(temp, "%s%s", base, INDEX_LOG);
	log_fn = (const char*)strcpy((char*)chkalloc(strlen(temp)+1,
						     sizeof(char)),temp);
#if OS_PCDOS			       /* open in binary mode (CR-LF) */
	if ((log_fp = OPEN_IN_B(log_fn)) == NULL) {
#else
	if ((log_fp = OPEN_IN(log_fn)) == NULL) {
#endif /* OS_PCDOS */
	    FATAL("Source log file %s not found.\n", log_fn);
	} else {
	    find_pageno();
	    CLOSE(log_fp);
	}
    }
#endif /* OS_BS2000 */
}


static void
find_pageno(VOID_ARG)
{
    int     i = 0;
    int     c;
    int     p;

#if (OS_VAXVMS | OS_BS2000 | OS_MVSXA)
    /*
     * Scan forward through the file for VMS, because fseek doesn't work on
     * variable record files
     */
    c = GET_CHAR(log_fp);
    while (c != EOF) {
	p = c;
	c = GET_CHAR(log_fp);
	if (p == LSQ && isdigit(c)) {
	    i = 0;
	    do {
		pageno[i++] = c;
		c = GET_CHAR(log_fp);
	    } while (isdigit(c));
	    pageno[i] = NUL;
	}
    }
    if (i == 0) {
	(void)fprintf(ilg_fp, "Couldn't find any page number in %s...ignored\n",
		log_fn);
	init_page = FALSE;
    }
#else  /* NOT (OS_VAXVMS | ...) */
    fseek(log_fp, -1L, 2);
    p = GET_CHAR(log_fp);
    fseek(log_fp, -2L, 1);
    do {
	c = p;
	p = GET_CHAR(log_fp);
    } while (!(((p == LSQ) && isdigit(c)) || (fseek(log_fp, -2L, 1) != 0)));
    if (p == LSQ) {
	while ((c = GET_CHAR(log_fp)) == SPC);
	do {
	    pageno[i++] = (char) c;
	    c = GET_CHAR(log_fp);
	} while (isdigit(c));
	pageno[i] = NUL;
    } else {
	(void)fprintf(ilg_fp, "Couldn't find any page number in %s...ignored\n",
		log_fn);
	init_page = FALSE;
    }
#endif /* (OS_VAXVMS | ...) */
}

static void
#if STDC
open_sty(char *fn)
#else
open_sty(fn)
char   *fn;
#endif
{
    char	*path;
    char	*ptr;
    int		i;
    int		len;
    STATIC char	temp[FILENAME_MAX];

#if OS_AMIGA /* suppress path search if the filename is absolute name */
    if (((path = getenv(STYLE_PATH)) == NULL) || strchr(fn,VOL_DELIM)
	|| strchr(fn,DIR_DELIM))
#else
    if ((path = getenv(STYLE_PATH)) == NULL)
#endif /* OS_AMIGA */
    {
	/* style input path not defined */
	strcpy(temp, fn);
	sty_fp = OPEN_IN(temp);
    } else {
#if OS_VAXVMS
	if ((strlen(STYLE_PATH) + strlen(fn) + 1) > sizeof(temp)) {
	    FATAL2("Path %s too long (max %u).\n", STYLE_PATH, 
		   (unsigned int)sizeof(temp));
	} else {
	    (void)sprintf(temp, "%s:%s", STYLE_PATH, fn);
	    sty_fp = OPEN_IN(temp);
	}
#else  /* NOT OS_VAXVMS */
	len = sizeof(temp) - (strlen(fn) + 1); /* space left in temp[] */
	while (*path != NUL) {
#if OS_XENIX
	    i = 0;
	    while ((*path != ENV_SEPAR) && (*path != NUL) && (i < len))
		temp[i++] = *path++;
#else  /* NOT OS_XENIX */
	    ptr = strchr(path, ENV_SEPAR);
	    i = 0;
	    if (ptr == (char*)NULL)
	    {
		int j = strlen(path);

		while (i < j)
		    temp[i++] = *path++;
	    }
	    else
	    {
		while ((path != ptr) && (i < len))
		    temp[i++] = *path++;
	    }
#endif /* OS_XENIX */
	    if (i == len) {
		FATAL2("Path %s too long (max %u).\n", temp, 
		       (unsigned int)sizeof(temp));
	    } else {
#if OS_VAXVMS
		if (temp[i-1] != ']')
		    temp[i++] = DIR_DELIM;
#else
#if OS_AMIGA
		if ((i == 1 && temp[0] == '.') || i == 0)
		{			/* '.' is the current directory */
		    temp[0] = NUL;
		}
		else /* Amiga paths don't need a DIR_DELIM after volume name */
		{
		    if ((temp[i-1] != VOL_DELIM) && (temp[i-1] != DIR_DELIM))
		        temp[i++] = DIR_DELIM;
		}
#else
		temp[i++] = DIR_DELIM;
#endif /* OS_AMIGA */
#endif
		temp[i] = NUL;
		strcat(temp, fn);
#if (OS_AMIGA || OS_XENIX)
		if (((sty_fp = OPEN_IN(temp)) == NULL) && (*path != NUL))
#else /* NOT (OS_AMIGA || OS_XENIX) */
		if ((sty_fp = OPEN_IN(temp)) == NULL)
#endif /* (OS_AMIGA || OS_XENIX) */
		    path++;
		else
		    break;
	    }
	}
#endif /* OS_VAXVMS */
    }

    if (sty_fp == NULL)
	FATAL("Index style file %s not found.\n", fn);
    sty_fn = (const char*)strcpy((char*)chkalloc(strlen(temp)+1,
						 sizeof(char)),temp);
}


page_number
#if STDC
strtopageno(char *str)
#else
strtopageno(str)
char   *str;
#endif
{
    page_number		val = 0;

    while (*str != NUL) {
	val = 10 * val + DIGITVALUE(*str);
	str++;
    }
    return (val);
}
