/***********************************************************************
Test the malloc() function to determine the overhead of successive
allocations of various sizes of memory.  The allocation chunk sizes to
be tested can be specified by command line arguments; if not given, a
single test is made with allocation size 1.

It is worthwhile to run this program with a wide range of allocation
sizes; the accompanying tstmall.sh script makes it easy to do so.

The behavior of some malloc() implementations, e.g. DEC ULTRIX 4.x and
MIPS EP/IX 2.1.1, is rather wasteful except for certain allocation
sizes.  A plot of the amount allocated versus the amount requested
should be a straight line, but on these systems, it is a stairstep
function with very large steps that reflect wastage of half the
memory.

[04-Apr-1994]
***********************************************************************/

#include <stdio.h>
#include <stdlib.h>

#if defined(__NeXT__)
void	free(void *);
void	*malloc(size_t);
#else
#include <malloc.h>
#endif

#ifndef EXIT_SUCCESS
#define EXIT_SUCCESS 0
#endif

#define MAXPRINT 32
#define MAXALLOC 10000

#if defined(__STDC__) || defined(__cplusplus)
#define STDC 1
#else
#define STDC 0
#endif

#if STDC
void	test(int);
#else
void	test();
#endif

#if STDC
int
main(int argc, char* argv[])
#else
int
main(argc,argv)
int argc;
char* argv[];
#endif
{
    int m;

    if (argc == 2)
       test(atoi(argv[1]));
    else
    {
	for (m = 1; m < argc; ++m)
	    test(atoi(argv[m]));
    }

    exit (EXIT_SUCCESS);
    return (EXIT_SUCCESS);
}


#if STDC
void
test(int howmuch)
#else
void
test(howmuch)
int howmuch;
#endif
{
    char *p = NULL;
    char *q = NULL;
    int k;
    int delta = 0;
    int total = 0;
    int n;

    printf("Test of successive malloc(%7d) calls\n",howmuch);
    p = (char*)malloc(howmuch);
    for (k = 1, n = 0 ; (k <= MAXALLOC) && (total <= 100000); ++k)
    {
	q = (char*)malloc(howmuch);
	total += (q - p);
	if ((delta != (int)(q - p)) && (n++ < MAXPRINT))
	    printf("%7d\tdelta = %7d -> %7d\ttotal = %7d\n",k,delta,(int)(q - p),
		   total);
	delta = q - p;
	p = q;
    }
    printf("%7d\tdelta = %7d\ttotal = %7d\n",k,delta,total);

    printf("\nTest of successive malloc(%7d)/free() calls\n",howmuch);
    p = (char*)malloc(howmuch);
    for (k = 1, n = 0 ; (n < MAXPRINT) && (k <= 100); ++k)
    {
	q = (char*)malloc(howmuch);
	if ((delta != (int)(q - p)) && (n++ < MAXPRINT))
	    printf("%7d\tdelta = %7d -> %7d\n",k,delta,(int)(q - p));
	delta = q - p;
	p = q;
	free(q);
    }
    printf("%7d\tdelta = %7d\n",k,delta);
}
